//! Asynchronous thread to check if result is returned (non-blocking)

use std::sync::{mpsc, RwLock};
use std::thread;

/// Asynchronous thread handle
pub struct AsyncThread<T: Send + Sync + 'static> {
    receiver: mpsc::Receiver<T>,
    result: RwLock<Option<T>>,
}

impl<T: Send + Sync + 'static> AsyncThread<T> {
    /// Spawns a thread
    ///
    /// # Arguments
    ///
    /// `f` is a function to be executed in a separate thread
    pub fn spawn<F>(f: F) -> AsyncThread<T>
    where
        F: FnOnce() -> T,
        F: Send + 'static,
    {
        let (sender, receiver) = mpsc::channel();

        thread::spawn(move || {
            if let Err(err) = sender.send((f)()) {
                panic!("sending failed! ({})", err);
            }
        });

        AsyncThread {
            receiver,
            result: RwLock::new(None),
        }
    }

    /// Checks if the thread is already returned
    /// Returns the Some(result) if the thread has finished its work,
    /// otherwise None
    pub fn check(&self) -> Result<bool, anyhow::Error> {
        let mut result = self
            .result
            .write()
            .map_err(|_err| anyhow::Error::msg("Asyncthread: Failed getting write lock"))?;

        match result.as_ref() {
            Some(_) => Ok(true),
            None => match self.receiver.try_recv() {
                Ok(res) => {
                    *result = Some(res);

                    Ok(true)
                }
                Err(_) => Ok(false),
            },
        }
    }

    /// consumes the result
    pub fn take(&self) -> Result<T, anyhow::Error> {
        let mut result = self
            .result
            .write()
            .map_err(|_err| anyhow::Error::msg("Asyncthread: Failed getting write lock"))?;

        if result.is_some() {
            // actually safe to not panic, since we just checked
            let res = result.take().unwrap();

            Ok(res)
        } else {
            Err(anyhow::Error::msg("Asyncthread: Result not present"))
        }
    }
}

impl<T: Send + Sync + 'static> AsyncThread<T>
where
    T: Clone,
{
    pub fn get(&self) -> Result<T, anyhow::Error> {
        let result = self
            .result
            .read()
            .map_err(|_err| anyhow::Error::msg("Asyncthread: Failed getting write lock"))?;

        match result.as_ref() {
            Some(res) => Ok(res.clone()),
            None => Err(anyhow::Error::msg("Asyncthread: Result not present")),
        }
    }
}
